#include <msp/core/maputils.h>
#include <msp/core/raii.h>
#include <msp/gltk/button.h>
#include <msp/gltk/label.h>
#include <msp/gltk/row.h>
#include "libr2c2/vehicle.h"
#include "controlpanel.h"
#include "engineer.h"
#include "placementghost.h"
#include "trainview.h"

using namespace std;
using namespace Msp;
using namespace R2C2;

ControlPanel::ControlPanel(Engineer &e, Train &t):
	engineer(e),
	train(t),
	updating(false),
	placing(false)
{
	Loader::WidgetMap widgets;
	DataFile::load(*this, "data/controlpanel.ui", widgets);

	dynamic_cast<GLtk::Label *>(get_item(widgets, "lbl_protocol"))->set_text(train.get_protocol());
	dynamic_cast<GLtk::Label *>(get_item(widgets, "lbl_address"))->set_text(lexical_cast<string>(train.get_address()));

	dynamic_cast<GLtk::Button *>(get_item(widgets, "btn_place"))->signal_clicked.connect(sigc::mem_fun(this, &ControlPanel::place_clicked));
	dynamic_cast<GLtk::Button *>(get_item(widgets, "btn_take"))->signal_clicked.connect(sigc::mem_fun(this, &ControlPanel::take_clicked));
	dynamic_cast<GLtk::Button *>(get_item(widgets, "btn_view"))->signal_clicked.connect(sigc::mem_fun(this, &ControlPanel::view_clicked));

	GLtk::Panel *pnl_functions = dynamic_cast<GLtk::Panel *>(get_item(widgets, "pnl_functions"));

	const VehicleType::FunctionMap &funcs = train.get_vehicle(0).get_type().get_functions();
	GLtk::Row row(*pnl_functions->get_layout());
	for(VehicleType::FunctionMap::const_iterator i=funcs.begin(); i!=funcs.end(); ++i)
	{
		GLtk::Toggle *tgl = new GLtk::Toggle(i->second);
		tgl->set_value(train.get_function(i->first));
		tgl->signal_toggled.connect(sigc::bind(sigc::mem_fun(this, &ControlPanel::ui_function_toggled), i->first));
		pnl_functions->add(*tgl);
		tgl_funcs[i->first] = tgl;
	}

	train.signal_function_changed.connect(sigc::mem_fun(this, &ControlPanel::train_function_changed));
}

void ControlPanel::ui_function_toggled(bool value, unsigned func)
{
	if(!updating)
		train.set_function(func, value);
}

void ControlPanel::train_function_changed(unsigned func, bool value)
{
	SetFlag setf(updating);
	map<unsigned, GLtk::Toggle *>::iterator i = tgl_funcs.find(func);
	if(i!=tgl_funcs.end())
		i->second->set_value(value);
}

void ControlPanel::place_clicked()
{
	signal_grab_pointer.emit();
	placing = true;

	Layout3D &layout3d = engineer.get_layout_3d();
	unsigned n_vehs = train.get_n_vehicles();
	for(unsigned i=0; i<n_vehs; ++i)
		ghosts.push_back(new PlacementGhost(layout3d, train.get_vehicle(i).get_type()));
}

void ControlPanel::take_clicked()
{
	train.unplace();
}

void ControlPanel::view_clicked()
{
	TrainView *dlg = new TrainView(engineer, train);
	find_ancestor<GLtk::Root>()->add(*dlg);
	dlg->autosize();
}

void ControlPanel::button_press(int x, int y, unsigned btn)
{
	Panel::button_press(x, y, btn);

	if(placing)
	{
		signal_ungrab_pointer.emit();
		placing = false;

		for(vector<PlacementGhost *>::iterator i=ghosts.begin(); i!=ghosts.end(); ++i)
			delete *i;
		ghosts.clear();

		if(btn==1 && place_location)
			train.place(place_location);
	}
}

void ControlPanel::pointer_motion(int x, int y)
{
	Panel::pointer_motion(x, y);

	if(placing)
	{
		int rx = x;
		int ry = y;
		map_coords_to_ancestor(rx, ry, *find_ancestor<GLtk::Root>());
		Ray ray = engineer.get_main_view().create_ray(rx, ry);
		Vector ground = ray.get_start()-ray.get_direction()*ray.get_start().z/ray.get_direction().z;
		Track *track = engineer.get_layout().pick<Track>(ray);
		if(track)
		{
			const vector<Block::Endpoint> &eps = track->get_block().get_endpoints();
			int closest_ep = -1;
			float closest_dist = -1;
			for(unsigned i=0; i<eps.size(); ++i)
			{
				Snap sn = eps[i].track->get_snap_node(eps[i].track_ep);
				float d = (sn.position-ground).norm();
				if(d<closest_dist || closest_dist<0)
				{
					closest_ep = i;
					closest_dist = d;
				}
			}

			if(closest_ep>=0)
			{
				place_location = BlockIter(&track->get_block(), closest_ep);
				ghosts.back()->place(place_location.track_iter());
				for(unsigned i=ghosts.size()-1; i--; )
					ghosts[i]->place_before(*ghosts[i+1]);
			}
		}
	}
}
