#ifndef LIBR2C2_VEHICLETYPE_H_
#define LIBR2C2_VEHICLETYPE_H_

#include <msp/datafile/objectloader.h>
#include "geometry.h"
#include "objecttype.h"

namespace R2C2 {

class VehicleType: public ObjectType
{
private:
	typedef std::map<std::string, unsigned> TagMap;

public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<VehicleType, ObjectType::Loader>
	{
	private:
		TagMap rod_tags;

	public:
		Loader(VehicleType &);
	private:
		virtual void finish();
		void axle();
		void bogie();
		void function(unsigned, const std::string &);
		void gauge(float);
		void height(float);
		void length(float);
		void mirror_rods();
		void rod(const std::string &);
		void width(float);
	};

	struct Bogie;

	struct Axle
	{
		class Loader: public Msp::DataFile::ObjectLoader<Axle>
		{
		public:
			Loader(Axle &);
		private:
			void position(float);
			void wheel_diameter(float);
		};

		unsigned index;
		Bogie *bogie;
		float local_position;
		float position;
		float wheel_dia;
		bool powered;
		std::string object;

		Axle();
	};

	struct Bogie
	{
		class Loader: public Msp::DataFile::ObjectLoader<Bogie>
		{
		private:
			VehicleType &parent;

		public:
			Loader(VehicleType &, Bogie &);
		private:
			void axle();
			void position(float);
		};

		unsigned index;
		float position;
		unsigned first_axle;
		std::vector<Axle *> axles;
		std::string object;
		bool rotate_object;

		Bogie();
	};

	struct RodConstraint
	{
		enum Type
		{
			MOVE,
			SLIDE,
			ROTATE
		};

		enum Target
		{
			BODY,
			BOGIE,
			AXLE,
			ROD
		};

		class Loader: public Msp::DataFile::ObjectLoader<RodConstraint>
		{
		private:
			TagMap &tags;

		public:
			Loader(RodConstraint &, TagMap &);
		private:
			void axis(float, float, float);
			void local_position(float, float, float);
			void target_axle(unsigned);
			void target_position(float, float, float);
			void target_rod(const std::string &);
		};

		Type type;
		Target target;
		unsigned target_index;
		Vector target_position;
		Vector local_position;
		Vector axis;

		RodConstraint();
	};

	struct Rod
	{
		class Loader: public Msp::DataFile::ObjectLoader<Rod>
		{
		private:
			TagMap &tags;

		public:
			Loader(Rod &, TagMap &);
		private:
			template<RodConstraint::Type>
			void constraint();
			void initial_position(float, float, float);
		};

		Vector initial_position;
		std::vector<RodConstraint> constraints;
		std::string object;
		bool mirror_object;

		Rod();
	};

	class MirrorParametersLoader: public Msp::DataFile::Loader
	{
	public:
		std::string filter;
		Angle phase_offset;

		MirrorParametersLoader();
	private:
		void filt(const std::string &);
		void phase_offs(float);
	};

	typedef std::vector<Axle> AxleArray;
	typedef std::vector<Bogie> BogieArray;
	typedef std::vector<Rod> RodArray;
	typedef std::map<unsigned, std::string> FunctionMap;

private:
	bool locomotive;
	FunctionMap functions;
	bool swap_direction;
	float gauge;
	float length;
	float width;
	float height;
	AxleArray axles;
	std::vector<Axle *> fixed_axles;
	BogieArray bogies;
	RodArray rods;
	std::string object;
	bool rotate_object;
	float max_speed;

public:
	VehicleType();

	bool is_locomotive() const { return locomotive; }
	unsigned get_max_function() const;
	const FunctionMap &get_functions() const { return functions; }
	bool get_swap_direction() const { return swap_direction; }
	float get_gauge() const { return gauge; }
	float get_length() const { return length; }
	float get_width() const { return width; }
	float get_height() const { return height; }
	const AxleArray &get_axles() const { return axles; }
	const Axle &get_axle(unsigned) const;
	const Axle &get_fixed_axle(unsigned) const;
	const BogieArray &get_bogies() const { return bogies; }
	const Bogie &get_bogie(unsigned) const;
	const Axle &get_bogie_axle(unsigned, unsigned) const;
	const RodArray &get_rods() const { return rods; }
	const Rod &get_rod(unsigned) const;
	float get_front_axle_offset() const;
	float get_back_axle_offset() const;
	const std::string &get_object() const { return object; }
	bool get_rotate_object() const { return rotate_object; }
	float get_maximum_speed() const { return max_speed; }
};

} // namespace R2C2

#endif
