#include <msp/core/refptr.h>
#include <msp/net/inet.h>
#include <msp/net/resolve.h>
#include "libr2c2/aicontrol.h"
#include "libr2c2/catalogue.h"
#include "libr2c2/driver.h"
#include "libr2c2/route.h"
#include "libr2c2/train.h"
#include "libr2c2/trainrouter.h"
#include "libr2c2/trainstatus.h"
#include "libr2c2/vehicletype.h"
#include "server.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Server::Server(Layout &l):
	layout(l),
	listen_sock(Net::INET),
	event_disp(0)
{
	layout.get_driver().signal_power.connect(sigc::hide(sigc::mem_fun(this, &Server::driver_state_changed)));
	layout.get_driver().signal_halt.connect(sigc::hide(sigc::mem_fun(this, &Server::driver_state_changed)));
	layout.signal_emergency.connect(sigc::mem_fun(this, &Server::emergency));

	layout.signal_train_added.connect(sigc::mem_fun(this, &Server::train_added));

	const map<unsigned, Train *> &trains = layout.get_trains();
	for(map<unsigned, Train *>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		train_added(*i->second);

	RefPtr<Net::SockAddr> addr = Net::resolve("*", "8315", Net::INET);
	listen_sock.listen(*addr, 4);
	listen_sock.signal_data_available.connect(sigc::mem_fun(this, &Server::incoming_connection));
}

Server::~Server()
{ }

void Server::use_event_dispatcher(IO::EventDispatcher &ed)
{
	event_disp = &ed;
	event_disp->add(listen_sock);
}

void Server::incoming_connection()
{
	Net::StreamSocket *sock = listen_sock.accept();
	if(event_disp)
		event_disp->add(*sock);
	connections.push_back(new Connection(*this, sock));
}

void Server::driver_state_changed()
{
	DriverStatePacket pkt;
	pkt.power = layout.get_driver().get_power();
	pkt.halt = layout.get_driver().is_halted();
	send(pkt);
}

void Server::emergency(Block *, const string &msg)
{
	EmergencyPacket pkt;
	pkt.message = msg;
	send(pkt);
}

void Server::train_added(Train &train)
{
	train.signal_control_changed.connect(sigc::bind<0>(sigc::mem_fun(this, &Server::train_control_changed), sigc::ref(train)));
	train.signal_function_changed.connect(sigc::bind<0>(sigc::mem_fun(this, &Server::train_function_changed), sigc::ref(train)));
	train.signal_ai_event.connect(sigc::bind<0>(sigc::mem_fun(this, &Server::train_ai_event), sigc::ref(train)));

	TrainInfoPacket pkt;
	pkt.address = train.get_address();
	pkt.loco_type = layout.get_catalogue().get_name(&train.get_locomotive_type());
	pkt.name = train.get_name();
	send(pkt);

	if(!train.get_ai_of_type<TrainStatus>())
		new TrainStatus(train);
}

void Server::train_control_changed(const Train &train, const string &control, float value)
{
	TrainControlPacket pkt;
	pkt.address = train.get_address();
	pkt.control = control;
	pkt.value = value;
	send(pkt);
}

void Server::train_function_changed(const Train &train, unsigned, bool)
{
	TrainFunctionPacket pkt;
	pkt.address = train.get_address();
	pkt.functions = train.get_functions();
	send(pkt);
}

void Server::train_ai_event(const Train &train, TrainAI &ai, const TrainAI::Message &ev)
{
	if(ev.type=="route-changed")
	{
		TrainRoutePacket pkt;
		pkt.address = train.get_address();
		if(const Route *route = ev.value.value<const Route *>())
			pkt.route = route->get_name();
		send(pkt);
	}
	else if(ev.type=="status-changed")
	{
		TrainStatusPacket pkt;
		pkt.address = train.get_address();
		pkt.status = ev.value.value<std::string>();
		send(pkt);
	}
	else if(ev.type=="target-speed-changed" || ev.type=="reverse-changed")
	{
		AIControl *control = dynamic_cast<AIControl *>(&ai);
		if(control)
		{
			TrainAIControlPacket pkt;
			pkt.address = train.get_address();
			pkt.target_speed = control->get_target_speed();
			pkt.reverse = control->get_reverse();
			send(pkt);
		}
	}
}

template<typename P>
void Server::send(const P &pkt)
{
	for(vector<Connection *>::const_iterator i=connections.begin(); i!=connections.end(); ++i)
		if(!(*i)->stale && (*i)->comm.is_handshake_done())
			(*i)->comm.send(pkt);
}


Server::Connection::Connection(Server &s, Net::StreamSocket *o):
	server(s),
	socket(o),
	comm(*socket, server.proto, *this)
{
	socket->signal_end_of_file.connect(sigc::mem_fun(this, &Connection::end_of_file));
	comm.signal_handshake_done.connect(sigc::mem_fun(this, &Connection::handshake_done));
	comm.signal_error.connect(sigc::mem_fun(this, &Connection::comm_error));
	comm.initiate_handshake();
}

Server::Connection::~Connection()
{
	delete socket;
}

void Server::Connection::handshake_done()
{
	{
		DriverStatePacket pkt;
		pkt.power = server.layout.get_driver().get_power();
		pkt.halt = server.layout.get_driver().is_halted();
		comm.send(pkt);
	}

	const set<Route *> &routes = server.layout.get_all<Route>();
	for(set<Route *>::const_iterator i=routes.begin(); i!=routes.end(); ++i)
		if(!(*i)->is_temporary())
		{
			RouteInfoPacket pkt;
			pkt.name = (*i)->get_name();
			comm.send(pkt);
		}

	const Catalogue &cat = server.layout.get_catalogue();
	const map<unsigned, Train *> &trains = server.layout.get_trains();
	for(map<unsigned, Train *>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
	{
		const Train &train = *i->second;

		{
			TrainInfoPacket pkt;
			pkt.address = train.get_address();
			pkt.loco_type = cat.get_name(&train.get_locomotive_type());
			pkt.name = train.get_name();
			comm.send(pkt);
		}
		if(AIControl *control = train.get_ai_of_type<AIControl>())
		{
			TrainAIControlPacket pkt;
			pkt.address = train.get_address();
			pkt.target_speed = control->get_target_speed();
			pkt.reverse = control->get_reverse();
			comm.send(pkt);
		}
		for(unsigned j=0;; ++j)
		{
			const char *name = train.get_controller().enumerate_controls(j);
			if(!name)
				break;

			TrainControlPacket pkt;
			pkt.address = train.get_address();
			pkt.control = name;
			pkt.value = train.get_control(name);
			comm.send(pkt);
		}
		{
			TrainFunctionPacket pkt;
			pkt.address = train.get_address();
			pkt.functions = train.get_functions();
			comm.send(pkt);
		}
		if(TrainStatus *status = train.get_ai_of_type<TrainStatus>())
		{
			TrainStatusPacket pkt;
			pkt.address = train.get_address();
			pkt.status = status->get_status();
			comm.send(pkt);
		}
		if(TrainRouter *router = train.get_ai_of_type<TrainRouter>())
		{
			TrainRoutePacket pkt;
			pkt.address = train.get_address();
			pkt.route = router->get_route()->get_name();
			comm.send(pkt);
		}
	}
}

void Server::Connection::end_of_file()
{
	stale = true;
}

void Server::Connection::receive(const DriverStatePacket &pkt)
{
	Driver &driver = server.layout.get_driver();
	if(pkt.power!=driver.get_power())
		driver.set_power(pkt.power);
	if(pkt.halt!=driver.is_halted())
		driver.halt(pkt.halt);
}

void Server::Connection::receive(const TrainControlPacket &pkt)
{
	try
	{
		Train &train = server.layout.get_train(pkt.address);
		train.set_control(pkt.control, pkt.value);
	}
	catch(const exception &e)
	{
		error(e.what());
	}
}

void Server::Connection::receive(const TrainFunctionPacket &pkt)
{
	try
	{
		Train &train = server.layout.get_train(pkt.address);
		for(unsigned i=0; i<9; ++i)
			if(((pkt.functions^train.get_functions())>>i)&1)
				train.set_function(i, (pkt.functions>>i)&1);
	}
	catch(const exception &e)
	{
		error(e.what());
	}
}

void Server::Connection::receive(const TrainAIControlPacket &pkt)
{
	try
	{
		Train &train = server.layout.get_train(pkt.address);
		train.ai_message(TrainAI::Message("set-target-speed", pkt.target_speed));
		train.ai_message(TrainAI::Message("set-reverse", pkt.reverse));
	}
	catch(const exception &e)
	{
		error(e.what());
	}
}

void Server::Connection::receive(const TrainRoutePacket &pkt)
{
	try
	{
		Train &train = server.layout.get_train(pkt.address);
		if(pkt.route.empty())
			train.ai_message(TrainAI::Message("clear-route"));
		else
		{
			Route &route = server.layout.get_route(pkt.route);
			train.ai_message(TrainAI::Message("set-route", &route));
		}
	}
	catch(const exception &e)
	{
		error(e.what());
	}
}

void Server::Connection::comm_error(const exception &)
{
	stale = true;
}

void Server::Connection::error(const string &msg)
{
	ErrorPacket pkt;
	pkt.message = msg;
	comm.send(pkt);
}

} // namespace R2C2
